#include <glib.h>
#include <gudev/gudev.h>

/*
 * Workaround for LP: #1794280.
 *
 * That bug is because the DRM device isn't ready by the time GDM tries to
 * start wayland/X.
 * This is a script to add to ExecStartPre of gdm.service. It does the
 * following:
 *
 * 1. Enumerate drm devices from udev, looking for a DRM master. If found,
 *    exit.
 * 2. Connect to the 'uevent' signal of gudev, watching for the same to be
 *    added. Again exit if any are found.
 * 3. If, after 10 seconds, we haven't seen anything, try to proceed anyway as
 *    a failsafe.
 */

static gboolean
handle_device (GUdevDevice *device)
{
        const gchar * const * tags;
        tags = g_udev_device_get_tags (device);
        g_debug ("%s\n", g_udev_device_get_name (device));
        if (g_strv_contains (tags, "master-of-seat"))
        {
                g_debug ("    is seat master\n");
                return TRUE;
        }

        return FALSE;
}

static void
uevent_cb (GUdevClient *client G_GNUC_UNUSED,
           gchar       *action,
           GUdevDevice *device,
           gpointer     user_data)
{
        GMainLoop *loop;

        g_debug ("uevent action: %s\n", action);

        loop = (GMainLoop *) user_data;

        if (g_strcmp0 (action, "add") == 0)
        {
                if (handle_device (device))
                {
                        g_debug ("        this is good\n");
                        g_main_loop_quit (loop);
                }
                else
                {
                        g_debug ("        this is bad\n");
                }
        }
}

int
main()
{
        const gchar * const subsystems[] = { "drm", NULL };

        g_autoptr(GList) devices = NULL;
        g_autoptr(GMainLoop) loop = NULL;
        g_autoptr(GUdevClient) udev_client = NULL;
        g_autoptr(GUdevEnumerator) enumerator = NULL;

        loop = g_main_loop_new (NULL, FALSE);

        udev_client = g_udev_client_new (subsystems);
        enumerator = g_udev_enumerator_new (udev_client);
        g_udev_enumerator_add_match_is_initialized (enumerator);
        g_udev_enumerator_add_match_subsystem (enumerator, "drm");

        devices = g_udev_enumerator_execute (enumerator);

        g_debug ("enumerating devices...\n");

        for (GList *l = devices; l != NULL; l = l->next)
        {
                g_autoptr(GUdevDevice) device = G_UDEV_DEVICE (l->data);

                if (handle_device (device))
                {
                        g_debug ("        good enough for gdm\n");
                        return EXIT_SUCCESS;
                }
        }

        g_signal_connect (udev_client, "uevent", G_CALLBACK (uevent_cb), loop);

        /* after 10 seconds we try anyway */
        g_timeout_add_seconds (10, (GSourceFunc) g_main_loop_quit, loop);

        g_main_loop_run (loop);

        return EXIT_SUCCESS;
}
