#include <glib.h>
#include <gudev/gudev.h>
#include <stdio.h>
#include <string.h>

/*
 * Workaround for LP: #1794280 and #1925344.
 *
 * That bug is because the DRM devices aren't ready by the time GDM tries to
 * start wayland/X.
 * This is a script to add to ExecStartPre of gdm.service. It does the
 * following:
 *
 * 1. Enumerate drm devices from udev, looking for card0 and/or card1.
 *    Depending on graphics mode, exit if one or both are found.
 * 2. Connect to the 'uevent' signal of gudev, watching for the same to be
 *    added. Again exit if any are found.
 * 3. If, after 10 seconds, we haven't seen anything, try to proceed anyway as
 *    a failsafe.
 */

static char *
freadln (char *path)
{
    FILE *file = fopen(path, "r");
    if (file == NULL) {
        return NULL;
    }

    char *line = NULL;
    size_t len = 0;
    getline(&line, &len, file);
    fclose(file);
    return line;
}

/*
 * Systems that have configured X for multiple devices, such as laptops with
 * muxless dual GPUs, require the DRM devices to be available before starting.
 */
static gboolean
requires_two_gpus (void)
{
    // FIXME: This only works for systems with NVIDIA cards, but Intel+AMD
    //        systems should also need it.
    gchar *mode = freadln("/etc/prime-discrete");

    gboolean dual_gpus = mode &&
        (strcmp(mode, "nvidia\n") == 0 ||
         strcmp(mode, "on-demand\n") == 0);

    free(mode);
    return dual_gpus;
}

static gboolean
is_expected_card(const gchar * const name)
{
    return name &&
        (strcmp(name, "card0") == 0 ||
         strcmp(name, "card1") == 0);
}

static void
uevent_cb (GUdevClient *client G_GNUC_UNUSED,
           gchar       *action,
           GUdevDevice *device,
           gpointer     user_data)
{
        GMainLoop *loop;

        g_debug ("uevent action: %s\n", action);

        loop = (GMainLoop *) user_data;

        if (g_strcmp0 (action, "add") == 0)
        {
                const gchar * name = g_udev_device_get_name (device);
                if (is_expected_card (name))
                {
                        g_debug ("        this is good\n");
                        g_main_loop_quit (loop);
                }
                else
                {
                        g_debug ("        this is bad\n");
                }
        }
}

int
main()
{
        const gchar * const subsystems[] = { "drm", NULL };
        const gboolean needs_two_gpus = requires_two_gpus();
        int nr_cards = 0;

        g_autoptr(GList) devices = NULL;
        g_autoptr(GMainLoop) loop = NULL;
        g_autoptr(GUdevClient) udev_client = NULL;
        g_autoptr(GUdevEnumerator) enumerator = NULL;

        loop = g_main_loop_new (NULL, FALSE);

        udev_client = g_udev_client_new (subsystems);
        enumerator = g_udev_enumerator_new (udev_client);
        g_udev_enumerator_add_match_is_initialized (enumerator);
        g_udev_enumerator_add_match_subsystem (enumerator, "drm");

        devices = g_udev_enumerator_execute (enumerator);

        g_debug ("enumerating devices...\n");

        for (GList *l = devices; l != NULL; l = l->next)
        {
                g_autoptr(GUdevDevice) device = G_UDEV_DEVICE (l->data);
                const gchar * name = g_udev_device_get_name (device);

                if (!is_expected_card (name))
                {
                        continue;
                }

                g_debug ("%s\n", name);
                nr_cards++;

                if (!needs_two_gpus || nr_cards == 2)
                {
                        g_debug ("        good enough for gdm\n");
                        return EXIT_SUCCESS;
                }
        }

        g_signal_connect (udev_client, "uevent", G_CALLBACK (uevent_cb), loop);

        /* after 10 seconds we try anyway */
        g_timeout_add_seconds (10, (GSourceFunc) g_main_loop_quit, loop);

        g_main_loop_run (loop);

        return EXIT_SUCCESS;
}
